import os
from pathlib import Path

from coast import BlastBatch, coast_var as coast_var


class DiamondBatch(BlastBatch):
    def __init__(self,
                 query,
                 db,
                 subject=None,
                 outfmt=None,
                 out="blast_results.tab",
                 taxonlist=None,
                 taxidlist=None,
                 sensitivity=None,
                 matrix=None,
                 evalue=None,
                 gapextend=None,
                 gapopen=None):
        """
        Formulate BLAST+ command line query.

        Parameters
        ----------
        query: str
            path for the file with the proteome, either GenBank or fasta.
        db: str
            Path to the BLAST database to be used.
        out: str
            Path to align results in tabular format.
        taxonlist: str
            Comma separated taxid values for positive filtering of the selected db.
        taxidlist: str
            Path to the file with the taxid values for positive filtering of the selected db.
        matrix: str
            Substitution matrix.
        evalue: decimal
            Evalue filter.
        gapextend: int
            Penalty for gap extension.
        gapopen: int
            Penalty for gap opening.
        """
        super().__init__(query=query, db=db, outfmt=outfmt, out=out, taxonlist=taxonlist,
                         matrix=matrix, evalue=evalue, gapextend=gapextend, gapopen=gapopen, subject=subject)

        self.engine_dict = coast_var.aln_options["diamond"]
        self.merlin_engine = "DiamondP"
        self.tool_command = "diamond blastp"
        self.sensitivity = sensitivity
        self.outfmt = f'6 {coast_var.output_format["diamond"]}'
        if taxidlist:
            with open(taxidlist) as taxidlist:
                taxidlist = taxidlist.readlines()
                if len(taxidlist) <= 500:
                    self.taxonlist = f"'{','.join([taxid.strip() for taxid in taxidlist.readlines()])}'"
                else:
                    raise Exception(
                        "The --taxonlist file provided includes too many sequences. You could identify higher order "
                        "common taxid nodes to reduce the necessary number of taxids.")

    def _check_db(self, db):
        if Path(f"{db}.dmnd").exists():
            return True
        else:
            return False

    def merlin_parameters(self):
        """
        Generates a merlin api call parameters.

        Returns
        -------
        str
            Parameters to be written to file.
        """
        query_parameters = []

        for option in ["db", "outfmt", "taxonlist", "matrix", "evalue", "gapopen", "gapextend"]:
            if getattr(self, option):
                query_parameters.append(f'{self.engine_dict[option].strip("--").strip(" ")} = {getattr(self, option)}')

        return query_parameters

    def run_merlin(self, email, address='https://rosalind.di.uminho.pt:6082/'):
        """
        Sends merlin api align request.

        Parameters
        ----------
        email: str
            Email used for request logging.
        address: str
            Api server base address for request.
        """
        protein_path = self.workspace / self.query
        parameters_file_path = self.workspace / "SubmissionParameters.txt"
        files = [protein_path, parameters_file_path]
        # Generate params
        with open(parameters_file_path, mode='wt', encoding='utf-8') as parameters_file:
            parameters_file.write('\n'.join(self.merlin_parameters()))

        self.request_merlin(files, address, email, raw_output="DiamondResults.out", command="DiamondP")
