import argparse
import os

from coast.cli.report import report_handler
from coast.cli.compare import compare_handler
from coast.cli.search import search_handler


def coast():
    # TODO: Add parent subparsers https://stackoverflow.com/questions/33645859/how-to-add-common-arguments-to
    #  -argparse-subcommands
    parser = argparse.ArgumentParser(
        description="coast - Comparative Omnics Alignment Search Tool v0.1"
    )
    # prints help when there is not subcommand provided
    parser.set_defaults(func=lambda args: parser.print_help())
    subparsers = parser.add_subparsers(title='subcommands',
                                       description='valid subcommands',
                                       help='additional help')

    # Reporting
    report = subparsers.add_parser(
        "report",
        description="Processes data generated in searchs or comparissions"
    )
    report.add_argument("results",
                        help="Path to the align results from the original coast run.",
                        type=str)
    report.add_argument("queries",
                        help="Produce the similarity matrices from a single or different GenBanks and FASTAs files. "
                             "Each represents an organism. Must be the same used in the coast search/ coast compare "
                             "query.",
                        type=str,
                        nargs='+')
    report.add_argument("-o", "--out",
                        help="The directory where the output will be delivered.",
                        type=str)
    report.add_argument("-f", "--outfmt",
                        help="Choose the output format that will be used for the produced data. It can either be the "
                             "tabular, human readable text file or full workspace with intermediary files and extra "
                             "results.",
                        type=str,
                        nargs='*',
                        default=['b', 'a', 'r'])
    report.add_argument("-k", "--keywords",
                        help="This option is only taken into account when a GenBank File is provided, in such cases, "
                             "it identifies the annotation keywords used to identify proteins, default is only using "
                             "CDS.",
                        type=str)
    report.add_argument("--aai",
                        help="Minimum Average Amino Acid Identity, for reporting purposes",
                        default=10,
                        type=int)
    report.add_argument("--id",
                        help="Minimum Amino Acid Identity, for hit selection for AAIbd calculation",
                        default=40,
                        type=int)
    report.add_argument("--cov",
                        help="Minimum coverage, for hit selection for AAIbd calculation",
                        default=50,
                        type=int)
    report.add_argument('--rbh', dest='reciprocal', action='store_true',
                        help="Analysis for reciprocal best hits."
                             "This should be used for 'coast compare' results"
                        )
    report.add_argument('--quiet', dest='verbose', action='store_false',
                        help="Deactivate verbose terminal output along the created files."
                             "This is already the default behaviour."
                        )
    report.add_argument('--verbose', dest='verbose', action='store_true',
                        help="Activate verbose terminal output along the created files."
                        )
    report.add_argument('--filter_hypothetical', dest='hypothetical', action='store_true',
                        help="Filters proteins with 'hypothetical' in their description, use with caution."
                        )
    report.set_defaults(func=report_handler, verbose=False, out=os.getcwd(), engine="blast", keywords="CDS", reciprocal=False,
                        hypothetical=False)

    # Search
    search = subparsers.add_parser("search")
    # TODO: add file validation
    # TODO: add proper subcomand descriptions
    search.add_argument("database",
                        help="Directory for the database against which the query will be run.",
                        type=str)
    search.add_argument("query",
                        help="Produce the similarity matrices from a GenBank or FASTA file.",
                        type=str)
    search.add_argument("-o", "--out",
                        help="The directory where the output will be delivered.",
                        type=str)
    search.add_argument("-e", "--engine",
                        choices=['blast', 'diamond'],
                        help="Alignment Engine.",
                        type=str)
    search.add_argument("-k", "--keywords",
                        help="This option is only taken into account when a GenBank File is provided, in such cases, "
                             "it identifies the annotation keywords used to identify proteins, default is only using "
                             "CDS.",
                        type=str)
    taxonomy = search.add_mutually_exclusive_group()
    taxonomy.add_argument("--taxidlist",
                          help="File with NCBI taxonomic IDs to filter the database by.",
                          type=str)
    taxonomy.add_argument("--taxonlist",
                          help="Comma-separated list of NCBI taxonomic IDs to filter the database by.",
                          type=str)
    engine_options = search.add_mutually_exclusive_group()
    engine_options.add_argument("--task",
                                choices=['blastp', 'blastp-fast', 'blastp-short'],
                                help="Setting for the blast task. Read the BLAST documentation for more information.",
                                type=str)
    engine_options.add_argument("--sens",
                                choices=['mid-sensitive', 'sensitive', 'more-sensitive', 'very-sensitive',
                                         'ultra-sensitive'],
                                help="Setting for the diamond align tool sensibility.",
                                type=str)
    search.add_argument("--matrix",
                        choices=['BLOSUM45', 'BLOSUM50', 'BLOSUM62', 'BLOSUM80', 'BLOSUM90', 'PAM250', 'PAM70',
                                 'PAM30'],
                        help="Scoring matrix",
                        type=str)
    search.add_argument("--evalue",
                        help="E-Value Threshold",
                        type=str)
    search.add_argument("--gapopen",
                        help="Gap open penalty.",
                        type=str)
    search.add_argument("--gapextend",
                        help="Gap extension penalty.",
                        type=str)
    search.add_argument("--threads",
                        help="Number of threads available for the engine to run. Read the respective engine's "
                             "documentation for further details.",
                        type=str)
    search.add_argument("-cont", "--containerization",
                        choices=['docker', 'podman'],
                        help="Containerization system, if not provided will run with local software. It must be OCI "
                             "complaint. Be sure that your user has docker execution "
                             "priviliges, so you dont need to provide sudo authorization.",
                        type=str,
                        default="docker")
    search.add_argument("-img", "--image",
                        help="Image for the container to run the selected engine. Be sure that the container contains "
                             "the actual software installed. If you dont provide a -cont argument, it will be run "
                             "with docker as default. If your system lacks docker it will error out",
                        type=str)
    search.add_argument("-w", "--web",
                        help="Merlin server address.",
                        default="http://rosalind.di.uminho.pt:6082/",
                        type=str)
    search.add_argument("--email",
                        help="Email for the Merlin server run.",
                        type=str)
    search.add_argument('--quiet', dest='verbose', action='store_false',
                        help="Deactivate verbose terminal output along the created files."
                             "This is already the default behaviour."
                        )
    search.add_argument('--verbose', dest='verbose', action='store_true',
                        help="Activate verbose terminal output along the created files."
                        )
    search.add_argument('--generate_workspace', dest='gen_ws', action='store_true',
                        help="Create the workspace inside a named and dated directory."
                        )
    search.add_argument('--inplace_workspace', dest='gen_ws', action='store_false',
                        help="Generate all files straight to the current working directory or the specified directory."
                        )
    search.set_defaults(func=search_handler, verbose=False, out=os.getcwd(), engine="blast", keywords="CDS",
                        gen_ws=True)

    compare = subparsers.add_parser("compare")
    # TODO: add file validation
    # TODO: add proper subcomand descriptions
    compare.add_argument("queries",
                         help="Produce the similarity matrices from different GenBanks and FASTAs files. Each "
                              "represents an organism.",
                         type=str,
                         nargs='+')
    compare.add_argument("-o", "--out",
                         help="The directory where the output will be delivered.",
                         type=str)
    compare.add_argument("-e", "--engine",
                         choices=['blast'],
                         help="Alignment Engine.",
                         type=str)
    compare.add_argument("-k", "--keywords",
                         help="This option is only taken into account when a GenBank File is provided, in such cases, "
                              "it identifies the annotation keywords used to identify proteins, default is only using "
                              "CDS.",
                         type=str)
    compare.add_argument("--task",
                         choices=['blastp', 'blastp-fast', 'blastp-short'],
                         help="Setting for the blast task. Read the BLAST documentation for more information.",
                         type=str)
    compare.add_argument("--matrix",
                         choices=['BLOSUM45', 'BLOSUM50', 'BLOSUM62', 'BLOSUM80', 'BLOSUM90', 'PAM250', 'PAM70',
                                  'PAM30'],
                         help="Scoring matrix",
                         type=str)
    compare.add_argument("--evalue",
                         help="E-Value Threshold",
                         type=str)
    compare.add_argument("--gapopen",
                         help="Gap open penalty.",
                         type=str)
    compare.add_argument("--gapextend",
                         help="Gap extension penalty.",
                         type=str)
    compare.add_argument("--threads",
                         help="Number of threads available for the engine to run. Read the respective engine's "
                              "documentation for further details.",
                         type=str)
    compare.add_argument("-cont", "--containerization",
                         choices=['docker', 'podman'],
                         help="Containerization system, if not provided will run with local software. It must be OCI "
                              "complaint. Be sure that your user has docker execution "
                              "priviliges, so you dont need to provide sudo authorization.",
                         type=str)
    compare.add_argument("-img", "--image",
                         help="Image for the container to run the selected engine. Be sure that the container contains "
                              "the actual software installed. If you dont provide a -cont argument, it will be run "
                              "with docker as default. If your system lacks docker it will error out",
                         type=str)
    compare.add_argument("-w", "--web",
                         help="Merlin server address.",
                         default="http://rosalind.di.uminho.pt:6082/",
                         type=str)
    compare.add_argument("--email",
                         help="Email for the Merlin server run.",
                         type=str)
    compare.add_argument('--quiet', dest='verbose', action='store_false',
                         help="Deactivate verbose terminal output along the created files."
                              "This is already the default behaviour."
                         )
    compare.add_argument('--verbose', dest='verbose', action='store_true',
                         help="Activate verbose terminal output along the created files."
                         )
    compare.add_argument('--generate_workspace', dest='gen_ws', action='store_true',
                         help="Create the workspace inside a named and dated directory."
                         )
    compare.add_argument('--inplace_workspace', dest='gen_ws', action='store_false',
                         help="Generate all files straight to the current working directory or the specified directory."
                         )
    compare.set_defaults(func=compare_handler, verbose=False, out=os.getcwd(), engine="blast", keywords="CDS",
                         gen_ws=True)

    args = parser.parse_args()
    try:
        args.func(args)
    except (Exception, ArithmeticError) as e:
        parser.error(e)


if __name__ == '__main__':
    coast()
