from Bio import SeqIO
from pathlib import Path
from coast import BlastBatch, DiamondBatch, COASTResults
from coast.util.input import loader
from coast.util.workspace import generate_folder


def search_handler(args):
    """Handles args coming from the 'coast compare' subcommand"""
    search(query=args.query,
           db=args.database,
           engine=args.engine,
           keywords=args.keywords,
           output_path=args.out,
           task=args.task,
           matrix=args.matrix,
           evalue=args.evalue,
           gapextend=args.gapextend,
           gapopen=args.gapopen,
           sensitivity=args.sens,
           taxonlist=args.taxonlist,
           taxidlist=args.taxidlist,
           containerization=args.containerization,
           image=args.image,
           email=args.email,
           web=args.web,
           threads=args.threads,
           gen_ws=args.gen_ws)

    # defaults
    if args.threads and args.image:
        raise Exception("Cannot run in local and container context.")
    if args.threads and args.email:
        raise Exception("Cannot run in local and web context.")
    if args.email and args.image:
        raise Exception("Cannot run in web and container context.")
    if args.task and args.engine != "blast":
        raise Exception("The blast task options are exclusive to the blast mode.")
    if args.sens and args.engine != "diamond":
        raise Exception("The diamond sensitivity options are exclusive to the diamond mode.")

    return args


def search(query,
           db,
           keywords="CDS",
           engine="blast",
           output_path=None,
           task=None,
           matrix=None,
           evalue=None,
           gapextend=None,
           gapopen=None,
           sensitivity=None,
           taxonlist=None,
           taxidlist=None,
           containerization=None,
           image=None,
           email=None,
           web=None,
           threads=None,
           gen_ws=None):
    # start_time = datetime.datetime.now()
    query = Path(query)
    organism = query.stem
    sequences = loader(str(query), keywords)
    wd = Path(output_path)

    if gen_ws:
        wd = generate_folder(str(wd.resolve()), organism)
    SeqIO.write(sequences.values(), wd.joinpath("protein.faa").resolve(), "fasta")

    if image and taxidlist:
        import shutil
        shutil.copy2(taxidlist, wd)

    # log.write(f"start workspace:{start_time}\n")
    # log.write(f"start align:{datetime.datetime.now()}\n")
    if engine == "blast":
        aln = BlastBatch(query=wd.joinpath("protein.faa").resolve(),
                         out="blast_results.tab",
                         db=db,
                         taxonlist=taxonlist,
                         taxidlist=taxidlist,
                         task=task,
                         matrix=matrix,
                         evalue=evalue,
                         gapextend=gapextend,
                         gapopen=gapopen)
    elif engine == "diamond":
        aln = DiamondBatch(query=wd.joinpath("protein.faa").resolve(),
                           out="blast_results.tab",
                           db=db,
                           taxonlist=taxonlist,
                           taxidlist=taxidlist,
                           sensitivity=sensitivity,
                           matrix=matrix,
                           evalue=evalue,
                           gapextend=gapextend,
                           gapopen=gapopen)
    else:
        raise Exception("Engine is not available or selected.")

    if image and aln:
        aln.run_container(containerization=containerization, image=image)
    elif email:
        aln.run_merlin(email=email, address=web)
    else:
        aln.run_local(threads=threads)

    # save coast formatted file
    organism_dict = {organism: list(sequences.keys())}
    b = COASTResults.from_alignment_file(wd.joinpath("blast_results.tab"), organism_dict)
    b.save_file(wd.joinpath("coast_aln.tab"))
