from copy import copy
import pandas as pd

from coast.data.alignment_results import AlignmentResults
from coast.coast_var import output_format


class COASTResults(AlignmentResults):
    def __init__(self, results, tax=False, organisms=None):
        super().__init__(results, tax)
        self.query_organisms = organisms

    @classmethod
    def from_alignment_results(cls, alignment_results, organisms: dict):
        """ Make it compatible with COAST files ( adds query organism data )
        """
        organisms_table = {"name": [], "seqid": []}
        for organism in organisms.keys():
            for protein in organisms[organism]:
                organisms_table["name"].append(organism)
                organisms_table["seqid"].append(protein)

        organisms = pd.DataFrame().from_dict(organisms_table)
        df = alignment_results.results.merge(organisms.rename(columns={"name": "qname", "seqid": "qseqid"}),
                                               how="left",
                                               on=["qseqid"])
        if not alignment_results.tax:
            df = df.merge(organisms.rename(columns={"name": "sname", "seqid": "sseqid"}), on=["sseqid"])

        return cls(df, alignment_results.tax, organisms)


    @classmethod
    def from_alignment_file(cls, file_path, organisms: dict, header=""):
        blast_result = AlignmentResults.from_file(file_path, header)
        return cls.from_alignment_results(blast_result, organisms)

    @classmethod
    def from_file(cls, file_path, header=""):
        """
        Loads tabular output from file

        Parameters
        ----------
        file_path: str
            path for the file, formatted in tabular 6 output, according to the column specification
        header: list of str
            column names for the tabular output, by order

        Returns
        -------
        COAST.BlastResults
            BlastResults object with the tabular data

        """
        if not header:
            header = output_format["blast"].split()
        results = pd.read_table(file_path, names=header, skiprows=[0])
        tax = False
        if "staxids" in header:
            tax = True
        return cls(results, tax)

    def reciprocal_best_hits(self, min_cov=0):
        """
        Filters the best hits in a blast id, aggregated by query sequence and taxid, designed to simplify similarity
        matrix intake.

        Returns
        -------
        COAST.BlastResults
            References self BlastResults object with the tabular data

        """
        _results = self.results
        if min_cov > 0:
            _results = self.results[self.results.qcovs > min_cov]

        if self.tax:
            raise Exception("Not Available for coast search results")
            # results = _results.sort_values('pident', ascending=False).drop_duplicates(['qseqid', 'staxids'])
        else:
            results = _results.sort_values('pident', ascending=False).drop_duplicates(['qname', 'qseqid', 'sname'])
            results = results.append(copy(results[results.qname == results.sname]))
            results["seqhash"] = results[['qseqid', 'sseqid']].apply(lambda x: hash(' '.join(sorted(x))), axis=1)
            results["namehash"] = results[['qname', 'sname']].apply(lambda x: hash(' '.join(sorted(x))), axis=1)
            non_reciprocal = results.duplicated(subset=['seqhash', 'namehash'], keep=False)
            results = results.loc[non_reciprocal, :].drop_duplicates(['qname', 'qseqid', 'sname'])
        return COASTResults.from_df(results)

    def count_hits(self):
        """
        Lists the hits per combination of qseqid, qname e sname

        Returns
        -------
        pandas.DataFrame
            Columns:
                qseqid: str
                staxids: str
                hits: int

        """
        if self.tax:
            return self.results.groupby(by=["qseqid", "qname", "staxids"], as_index=False).size()
        else:
            return self.results.groupby(by=["qseqid", "qname", "sname"], as_index=False).size()

    def get_query(self):
        """
        Lists the different query sequences and the respective organism

        Returns
        -------
        pandas.DataFrame
            Columns:
                qseqid: str
                staxids: str
                hits: int

        """
        return self.results.loc[:, ["qname", "qseqid"]].copy().drop_duplicates(["qname", "qseqid"])
