from datetime import datetime
from pathlib import Path
from copy import copy

import pandas as pd

from coast.util.input import dataframe_from_sequences
from coast.data.alignment_results import AlignmentResults


class SimilarityMatrix:
    def __init__(self, aln, query, organisms=None, min_cov=50, min_id=40, aggregation_keys=None):
        """
        """
        if aggregation_keys is None:
            aggregation_keys = ["staxids"]
        if organisms is None:
            organisms = []
        self.tabular_sm = None
        self.tabular_agr_sm = None
        self.aln = aln
        self.bh = aln.best_hits_from_blast()
        self.query = query
        self.query_count = len(query)
        self.qorganisms = organisms
        self.sorganisms = self.bh.get_organisms()
        self.aggregation_keys = aggregation_keys

        self.min_cov, self.min_id = None, None
        self.aggregated_metrics(min_cov=min_cov, min_id=min_id)

    @classmethod
    def from_results(cls, sub_sequences, results_path, organisms=[], min_cov=50, min_id=40):
        """ Generates a SimilarityMatrix object from a align file.
        Parameters
        ----------
        min_id: int
            Minimum Amino Acid Identity, for hit selection for AAIbd calculation.
        min_cov: int
            Minimum coverage, for hit selection for AAIbd calculation.
        sub_sequences: dict of Bio.SeqRecord
            Sequences of the reference proteome.
        results_path: str
            Path to align results in tabular format.
        organisms: str
            Identifier of the reference proteome to be used for reporting purposes.
        """
        aln = AlignmentResults.from_file(results_path)
        query = dataframe_from_sequences(sub_sequences)
        organisms = []
        organisms.extend(organisms)

        return cls(aln, query, organisms, min_id=min_id, min_cov=min_cov)

    def cumulative_identification(self):
        """
        Through a left join, it identifies for each taxid in the bh dataframe, and adds, the proteins that are missing
        from the query.

        Returns
        -------
        pandas.Dataframe
            Similarity matrix in a tabular format
        """
        targets = self.query.merge(pd.DataFrame(data=self.sorganisms, columns=["staxids"]), how="cross")
        self.tabular_sm = targets.merge(self.bh.results, how="left", on=["qseqid", "staxids"]).sort_values(
            by=['qseqid', 'pident'])
        self.tabular_sm.pident = self.tabular_sm.pident.fillna(0)
        self.tabular_sm.qcovs = self.tabular_sm.qcovs.fillna(0)
        self.tabular_agr_sm = self.tabular_sm[["staxids", "sscinames", "sskingdoms"]].dropna().drop_duplicates()
        self.tabular_agr_sm["qcount"] = self.query_count
        return self.tabular_sm

    def __aaic(self):
        # Initiate aggregated df and calculates unweighted AAIc
        aaic = self.tabular_sm.groupby(by=self.aggregation_keys, as_index=False).agg(aaic=("pident", "sum"))
        self.tabular_agr_sm = self.tabular_agr_sm.merge(aaic, how='left', on=self.aggregation_keys)
        self.tabular_agr_sm.aaic = round(self.tabular_agr_sm.aaic / self.tabular_agr_sm.qcount, ndigits=3)

    def __aaibd(self, min_cov=50, min_id=40):
        # AAIb/AAId
        if min_cov != 0:
            bh_cov = self.aln.best_hits_from_blast(min_cov=min_cov)
            tabular_sm_cov = bh_cov.results
            candidate_hits = tabular_sm_cov[tabular_sm_cov.pident >= min_id].copy()
        else:
            candidate_hits = self.tabular_sm[self.tabular_sm.pident > self.min_id].copy()
        aaibd = candidate_hits.groupby(by=self.aggregation_keys, as_index=False).agg(aaibd=("pident", "mean"))
        self.tabular_agr_sm = self.tabular_agr_sm.merge(aaibd, how='left', on=self.aggregation_keys)
        self.tabular_agr_sm.aaibd = round(self.tabular_agr_sm.aaibd.fillna(0), ndigits=3)

    def __hitspp(self):
        # Hits per protein
        count = self.aln.count_hits().groupby(by=self.aggregation_keys, as_index=False).agg(hitspp=("size", "sum"))
        self.tabular_agr_sm = self.tabular_agr_sm.merge(count, how='left', on=self.aggregation_keys)
        self.tabular_agr_sm["hitspp"] = round(self.tabular_agr_sm.hitspp / self.tabular_agr_sm.qcount, ndigits=3)

    def __nmatch(self):
        # nº of match
        count = self.tabular_sm.groupby(by=self.aggregation_keys, as_index=False).agg(match=("nident", "count"))
        self.tabular_agr_sm = self.tabular_agr_sm.merge(count, how='left', on=self.aggregation_keys).sort_values(
            by=['aaic'], ascending=[0])
        self.tabular_agr_sm["pmatch"] = round(self.tabular_agr_sm.match / self.tabular_agr_sm.qcount, ndigits=3)

    def aggregated_metrics(self, min_cov=50, min_id=40):
        """
        Calculate the aggregated metrics (aai, hitspp, etc) for the similarity matrix, aggregated by organism (staxids).

        Returns
        -------
        pandas.Dataframe
            Aggregated metrics for the similarity matrix
        """
        if not isinstance(self.tabular_sm, pd.DataFrame):
            # To check dataframe the normal method is not compatible
            self.cumulative_identification()
        if isinstance(self.tabular_sm, pd.DataFrame):
            if self.tabular_sm.empty:
                self.cumulative_identification()

        self.min_cov, self.min_id = min_cov, min_id
        self.__aaic()
        self.__aaibd(min_id=self.min_id, min_cov=self.min_cov)
        self.__hitspp()
        self.__nmatch()
        return self.tabular_agr_sm

    def filtered_agr_sm(self, min_aai):
        """
        Filtered version of the agregated metrics table, filtered by minimum AAI score.

        Parameters
        ----------
        min_aai: int
             Minimum aai score for filtering in the report.

        Returns
        -------
        pandas.Dataframe
            Aggregated metrics for the similarity matrix.
        """
        return self.tabular_agr_sm[(self.tabular_agr_sm['aaic'] >= min_aai)].sort_values(by=['aaic', 'hitspp'],
                                                                                         ascending=False)

    def draw_aai_heatmap(self, path, min_aai=10):
        import matplotlib.pyplot as plt
        import seaborn as sns

        heatmap_path = Path(path) / "heatmap_plot.png"
        df = self.tabular_sm.set_index(['staxids', 'qseqid'])['pident'].unstack().reset_index()
        df = df.merge(self.filtered_agr_sm(min_aai=min_aai), on=["staxids"])
        df = df.set_index('staxids').sort_values(by=['aaic'], ascending=[0])
        height = len(df.aaic)
        width = self.query_count
        ratio = round(height / width, ndigits=1)

        df = df.drop(labels=["sscinames", "sskingdoms", "aaic", "aaibd", "hitspp", "match", "qcount", "pmatch"], axis=1)
        sns.set_theme(style="white")
        f, ax = plt.subplots(figsize=(4, 4 + ratio * 2))
        sns.heatmap(df, vmin=0, vmax=100, cmap="YlGnBu", square=False, cbar_kws={"shrink": .5})

        plt.subplots_adjust(left=0, top=1, bottom=0)
        plt.savefig(heatmap_path)
        return heatmap_path

    def draw_aai_interactive_heatmap(self, min_aai=10):
        from coast.graphing.heatmap import interactive_heatmap

        df = copy(self.tabular_sm[['qseqid', 'staxids', 'pident', 'qcovs']])
        mask = self.filtered_agr_sm(min_aai=min_aai)["staxids"],
        df = df.merge(mask[0], how="right", on="staxids")
        return interactive_heatmap(df, list(mask[0])[::-1], list(self.query["qseqid"]), title="AAIc Heatmap")

    def markdown_agr_sm(self, output_path, min_aai=10):
        """
        Produces a markdown report of the results.

        Parameters
        ----------
        output_path: str
            Path, including file name for the report to be saved at.

        min_aai: int
            Minimum aai score for filtering in the report.

        Returns
        ----------
        bool
            Execution confirmation
        """
        output_path = Path(output_path)
        heatmap_path = self.draw_aai_heatmap(path=output_path.parent, min_aai=min_aai)
        with open(output_path, "w") as report:
            report.write("# coast Results" + "\n" + "\n")
            report.write("## Summary" + "\n" + "\n")
            report.write(self.filtered_agr_sm(min_aai).to_markdown(index=False, floatfmt=".3f") + "\n" + "\n")
            report.write("## AAI Heatmap and Protein Association" + "\n" + "\n")
            report.write(f'![heatmap]({heatmap_path.name})')
        return output_path

    def html_agr_sm(self, output_path, min_aai=10):
        """
        Produces a html report of the results.

        Parameters
        ----------
        output_path: str
            Path, including file name for the report to be saved at.

        min_aai: int
            Minimum aai score for filtering in the report.

        Returns
        ----------
        bool
            Execution confirmation
        """
        from coast.util.report import reporting_env

        data = self.filtered_agr_sm(min_aai)
        if data.shape[0] != 0:
            # Generate heatmap
            heatmap = self.draw_aai_interactive_heatmap(min_aai)

            # generate table data
            table = [self.tabular_agr_sm.columns]
            for index, rows in self.filtered_agr_sm(min_aai).iterrows():
                # append the list to the final list
                table.append(rows)

            results = [
                {"name": "Table Summary",
                 "type": "table",
                 "content": table
                 },
                {"name": "AAIc Heatmap",
                 "type": "bokeh",
                 "content": heatmap
                 }
            ]
        else:
            results = [
                {"name": "Table Summary",
                 "type": "text",
                 "content": ["WARNING: No hits have been returned with the current AAIc filtering settings."]
                 },
                {"name": "AAIc Heatmap",
                 "type": "text",
                 "content": ["WARNING: No hits have been returned with the current AAIc filtering settings."]
                 }
            ]

        template = reporting_env.get_template("coast_report.html")
        render = template.render(title="COAST Report",
                                 min_aai=min_aai,
                                 min_id=self.min_id,
                                 min_cov=self.min_cov,
                                 run_time=datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                                 results=results)

        with open(output_path, "w") as report_file:
            report_file.write(render)
        return True

    def print_agr_sm(self, min_aai=10):
        """
        Prints command line tabulated summary of the aggregated metrics.

        Parameters
        ----------
        min_aai: int
            Minimum aai score for filtering in the report.
        """
        from tabulate import tabulate
        print(tabulate(self.filtered_agr_sm(min_aai), showindex=False, headers='keys', tablefmt='psql', floatfmt=".3f"))
