from datetime import datetime
from pathlib import Path

import numpy as np
import pandas as pd

from coast.coast_var import output_format
from coast.util.input import loader
from coast.sm import SimilarityMatrix
from coast import COASTResults


class ReciprocalSimilarityMatrix(SimilarityMatrix):
    def __init__(self, aln, query, organisms, min_cov=50, min_id=40, aggregation_keys=["qname", "sname"]):
        self.tabular_sm = None
        self.tabular_agr_sm = None
        self.aln = aln
        self.bh = aln.reciprocal_best_hits()
        self.query = query
        self.query_count = self.query.groupby(by=["qname"], as_index=False).agg(qcount=("qseqid", "nunique"))
        self.qorganisms = organisms
        self.sorganisms = list(organisms.keys())
        self.aggregation_keys = aggregation_keys

        self.min_cov, self.min_id = None, None
        self.aggregated_metrics(min_cov=min_cov, min_id=min_id)

    @classmethod
    def from_results(cls, sub_sequences, results_path, organisms=(), min_cov=50, min_id=40):
        ba = output_format["blast-short"].split()
        ba.extend(["qname", "sname"])
        aln = COASTResults.from_file(results_path, header=ba)

        organisms_table = {"qname": [], "qseqid": []}
        for organism in organisms.keys():
            for protein in organisms[organism]:
                organisms_table["qname"].append(organism)
                organisms_table["qseqid"].append(protein)
        query = pd.DataFrame.from_dict(organisms_table)

        return cls(aln, query, organisms, min_id=min_id, min_cov=min_cov)

    def cumulative_identification(self, min_cov=0):
        """
        Through a left join, it identifies for each taxid in the bh dataframe, and adds, the proteins that are missing
        from the query.

        Returns
        -------
        pandas.Dataframe
            Similarity matrix in a tabular format
        """
        targets = self.query.merge(pd.DataFrame(data=self.sorganisms, columns=["sname"]), how="cross")
        targets["namehash"] = targets[['qname', 'sname']].apply(lambda x: hash(' '.join(sorted(x))), axis=1)
        self.tabular_sm = targets.merge(self.bh.results, how="left",
                                        on=["qseqid", "sname", "qname", "namehash"]).sort_values(
            by=['qseqid', 'pident'])
        self.tabular_agr_sm = self.tabular_sm[self.aggregation_keys].dropna().drop_duplicates()
        self.tabular_agr_sm = self.tabular_agr_sm.merge(self.query_count, how="left", on=["qname"])
        return self.tabular_sm

    def similarity_matrix(self, data):
        """ transforms a tabular matrix into a squared one
        """
        if data in self.tabular_agr_sm.columns.values.tolist():
            _tabular = self.tabular_agr_sm.sort_values(by=['qname', 'sname'])
            _list = _tabular[data].to_list()
            org_num = len(self.qorganisms)
            org_name = sorted(list(self.qorganisms.keys()))
            array = np.array([_list[i * org_num:(i + 1) * org_num] for i in range(org_num)])
            return pd.DataFrame(data=array,
                                index=org_name,
                                columns=org_name)
        else:
            raise Exception("Data not in similarity matrix")

    def aaic(self):
        return self.similarity_matrix("aaic")

    def aaibd(self):
        return self.similarity_matrix("aaibd")

    def draw_aai_heatmap(self, path, min_aai=10):
        import matplotlib.pyplot as plt
        import seaborn as sns
        # import mpld3
        # from mpld3 import plugins

        heatmap_path = Path(path) / "heatmap_plot.png"
        plt.figure(figsize=(14, 7))
        sns.set_theme(style="white")
        sns.heatmap(data=self.aaibd(), vmin=0, vmax=100, cmap="YlGnBu")
        plt.savefig(heatmap_path)

        return heatmap_path

    def draw_aai_interactive_correlation_matrix(self, index="aaibd"):
        from coast.graphing.heatmap import interactive_correlation_matrix

        corr = self.tabular_agr_sm
        return interactive_correlation_matrix(corr, list(self.sorganisms)[::-1], list(self.sorganisms), title="AAIBD Heatmap", index=index)

    def draw_upgma(self):
        # reference http: // www.deduveinstitute.be / ~opperd / private / upgma.html
        pass

    def html_agr_sm(self, output_path):
        """
        Produces a html report of the results.

        Parameters
        ----------
        output_path: str
            Path, including file name for the report to be saved at.

        Returns
        ----------
        bool
            Execution confirmation
        """
        from coast.util.report import reporting_env

        # Generate heatmap
        heatmap_aaibd = self.draw_aai_interactive_correlation_matrix(index="aaibd")
        heatmap_aaic = self.draw_aai_interactive_correlation_matrix(index="aaic")

        # generate table data - AAIBD
        aaibd = self.aaibd()
        header = ["Organisms"]
        header.extend(aaibd.columns)
        aaibd_table = [header]
        for index, rows in aaibd.iterrows():
            # append the list to the final list
            row = [index]
            row.extend(rows)
            aaibd_table.append(row)

        # generate table data - AAIC
        aaic = self.aaic()
        header = ["Organisms"]
        header.extend(aaic.columns)
        aaic_table = [header]
        for index, rows in aaic.iterrows():
            # append the list to the final list
            row = [index]
            row.extend(rows)
            aaic_table.append(row)

        results = [
            {"name": "AAIBD Summary",
             "type": "table",
             "content": aaibd_table
             },
            {"name": "AAIBD Heatmap",
             "type": "bokeh",
             "content": heatmap_aaibd
             },
            {"name": "AAIC Summary",
             "type": "table",
             "content": aaic_table
             },
            {"name": "AAIC Heatmap",
             "type": "bokeh",
             "content": heatmap_aaic
             }
        ]

        template = reporting_env.get_template("coast_report.html")
        render = template.render(title="COAST Report",
                                 min_aai=None,
                                 min_id=self.min_id,
                                 min_cov=self.min_cov,
                                 run_time=datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                                 results=results)

        with open(output_path, "w") as report_file:
            report_file.write(render)

        return True

    def print_agr_sm(self, min_aai=10):
        """
        Prints command line tabulated summary of the aggregated metrics.

        Parameters
        ----------
        min_aai: int
            Minimum aai score for filtering in the report.
        """
        from tabulate import tabulate
        print(tabulate(self.filtered_agr_sm(min_aai), showindex=False, headers='keys', tablefmt='psql', floatfmt=".3f"))


if __name__ == "__main__":
    queries = ["sample_data/MN908947.3.gb", "sample_data/MN844877.gb"]
    sequences = {}
    organism_dict = {}

    for query in queries:
        query_path = Path(query)
        content = loader(str(query_path.resolve()))
        organism_dict[query_path.stem] = list(content.keys())
        sequences.update(loader(str(query_path.resolve())))

    path = Path("sample_data/coast_aln/coast_aln.tab")
    a = ReciprocalSimilarityMatrix.from_results(sub_sequences=sequences, results_path=path, organisms=organism_dict)
    b = a.aaibd()
    c = a.aaic()
    a.draw_aai_heatmap(str(Path.home()/"Desktop"))
    a.html_agr_sm(str(Path.home()/"Desktop"/"coast_report.html"))
