from pathlib import Path

from Bio import SeqIO
from Bio.Seq import Seq
import Bio.SeqIO
from Bio.SeqRecord import SeqRecord
import pandas as pd
import re


def sniffer(filename):
    """ Is it GENBANK, is it FASTA? I will smell it
    """
    header = open(filename).read(5)
    if header == 'LOCUS':
        return 'genbank'
    header = open(filename).read(1)
    if header == '>':
        return 'fasta'
    else:
        return ''


def proteins_from_genbank(genome, type="CDS", hypothetical=False):
    """
    Takes as input a SeqIO object created from a genebank, and extracts a list of proteins based in a feature key.

    Parameters
    ----------
    hypothetical: bool
        Boolean value to determine if hypothetical proteins should be filtered out.
    genome: Bio.SeqRecord
        SeqRecord object obtained from using SeqIO to read a GenBank file.
    type: str, default "CDS"
        Type of the features to be extracted from the Bio.SeqRecord feature set as proteins.

    Returns
    -------
    dict of str: Bio.SeqRecord
        returns a dict of proteins present in the file
    """
    proteins = {}
    if hypothetical:
        for feature in genome.features:
            if feature.type == type and not re.search("hypothetical", feature.qualifiers["product"][0]):
                proteins[feature.qualifiers["protein_id"][0]] = SeqRecord(
                    Seq(feature.qualifiers["translation"][0]),
                    id=feature.qualifiers["protein_id"][0],
                    description=feature.qualifiers["product"][0])
    else:
        for feature in genome.features:
            if feature.type == type:
                proteins[feature.qualifiers["protein_id"][0]] = SeqRecord(
                    Seq(feature.qualifiers["translation"][0]),
                    id=feature.qualifiers["protein_id"][0],
                    description=feature.qualifiers["product"][0])
    # the usage of [0] is because of the way biopython creates lists for many of this parameters even
    # though most times they only hold a single value, "THE" value
    return proteins


def proteins_from_fasta(genome, hypothetical=False):
    """
    Takes as input a SeqIO object created from a fasta, and extracts a list of proteins.

    Parameters
    ----------
    hypothetical: bool
        Boolean value to determine if hypothetical proteins should be filtered out.
    genome: Bio.SeqRecord
        SeqRecord object obtained from using SeqIO to read a GenBank file.

    Returns
    -------
    dict of str: Bio.SeqRecord
        returns a dict of proteins present in the file
    """
    proteins = {}
    if hypothetical:
        for protein in genome:
            if not re.search("hypothetical", protein.description):
                proteins[protein.id] = SeqRecord(
                    protein.seq,
                    id=protein.id,
                    description=protein.description)
    else:
        for protein in genome:
            proteins[protein.id] = SeqRecord(
                protein.seq,
                id=protein.id,
                description=protein.description)
    return proteins


def loader(filename, key="CDS", hypothetical=False):
    filename = Path(filename)
    format = sniffer(filename)
    file_ext = filename.suffix
    if format == "genbank":
        return proteins_from_genbank(SeqIO.read(filename, format=format), key, hypothetical=hypothetical)
    elif format == "fasta":
        return proteins_from_fasta(SeqIO.parse(filename, format=format), hypothetical=hypothetical)
    else:
        raise Exception(f'File "{file_ext}" format not recognized for the desired mode.')


def dataframe_from_sequences(sequences):
    """
    This is a sequence type agnostic function that will convert into a pandas dataframe with the relevant information
    coast needs.

    Parameters
    ----------
    sequences: list of Bio.SeqRecord or dict of str: Bio.SeqRecord

    Returns
    -------
    pandas.DataFrame
        Columns:
            rel_pos: int
            seq: str
            len: int

    """
    columns = ["rel_pos", "qseqid", "qlength"]
    _df = {col: [] for col in columns}
    rel_pos = 0
    for seq in sequences.values():
        rel_pos += 1
        _df["rel_pos"].append(rel_pos)
        _df["qseqid"].append(seq.id)
        _df["qlength"].append(len(seq))
    return pd.DataFrame(data=_df, columns=columns)



# def fastaer_generator(file):
#     """ FASTAer files are a file based index/metadata arquive meant to provide a stable querible data structure to
#     dead with the crazyness of dealing with FASTA and genbank inputs
#
#     quser_name: {qtaxids:[],
#             qtitle:"",
#             qscinames:[],
#             qtitle:[],
#             qseqs:[]
#             }
#     """
