import subprocess
import os
from pathlib import Path

import coast.coast_var as coast_var
import requests
import time
import json
from urllib import request
from shutil import unpack_archive
import hashlib


class BlastBatch:
    def __init__(self,
                 query,
                 db="",
                 subject="",
                 outfmt="",
                 out="blast_results.tab",
                 taxonlist="",
                 taxidlist="",
                 task="",
                 matrix="",
                 evalue="",
                 gapextend="",
                 gapopen=""):
        """
        Formulate BLAST+ command line query.

        Parameters
        ----------
        query: str
            path for the file with the proteome, either GenBank or fasta.
        db: str
            Path to the BLAST database to be used.
        subject:
            Path for the file with the the subject proteins, fasta.
        out: str
            Path to align results in tabular format.
        taxonlist: str
            Comma separated taxid values for positive filtering of the selected db.
        taxidlist: str
            Path to the file with the taxid values for positive filtering of the selected db.
        task: str
            BLASTP task name blasp, fast-blastp, etc ...
        matrix: str
            Substitution matrix.
        evalue: decimal
            Evalue filter.
        gapextend: int
            Penalty for gap extension.
        gapopen: int
            Penalty for gap opening.
        """
        self.engine_dict = coast_var.aln_options["blast"]
        self.merlin_engine = "BlastP"
        self.threads = ""
        self.tool_command = "blastp"
        self.subject = subject
        self.query = Path(query).name
        self.db = db
        self.taxidlist = taxidlist
        self.taxonlist = taxonlist
        self.task = task
        self.matrix = matrix
        self.evalue = evalue
        self.gapextend = gapextend
        self.gapopen = gapopen
        if not outfmt:
            self.outfmt = f'"6 {coast_var.output_format["blast"]}"'
        else:
            self.outfmt = outfmt
        self.out = out
        self.workspace = Path(query).parent

    def generic_aligner(self, db="", query="", out="", taxidlist=""):
        """
        Generates generic align command for BLAST and DIAMOND tools.
        if not provided special folders, it will use the workspace

        Parameters
        ----------
        db: str
            Path for the align database.
        query: str
            Path for the query file.
        out: str
            Path for the align result file.
        taxidlist: str
            Path for the taxidlist filter file.

        Returns
        -------
        str
            Alignment run command.
        """
        # TODO: Refactor
        _query, _out = self.query, self.out
        self.query, self.out = query, out

        if not self.db and db:
            _subject = self.subject
            self.subject = db
        elif db:
            _db = self.db
            self.db = db

        if taxidlist:
            _taxidlist = self.taxidlist
            self.taxidlist = taxidlist

        for option in self.engine_dict.keys():
            if getattr(self, option):
                self.tool_command = " ".join([self.tool_command, f'{self.engine_dict[option]}{getattr(self, option)}'])

        self.query, self.out = _query, _out

        if taxidlist:
            self.taxidlist = _taxidlist

        if not self.db and db:
            self.subject = _subject
        elif db:
            self.db = _db

        return self.tool_command

    def run_local(self, threads):
        """
        Runs batch align locally.

        Parameters
        ----------
        threads: int
            Number of threads to run the align tool.
        """
        self.threads = threads
        try:
            subprocess.run(self.generic_aligner(db=self.db,
                                                query=self.workspace / self.query,
                                                out=self.workspace / self.out),
                           shell=True,
                           capture_output=True,
                           check=True)
        except subprocess.CalledProcessError as e:
            raise Exception(e.stderr.decode('UTF-8'))

    def _check_db(self, db):
        db = Path(db)
        if db.with_suffix(".pdb").exists():
            return True
        elif db.with_suffix(".faa").exists():
            return True
        elif db.with_suffix(".fa").exists():
            return True
        elif db.with_suffix(".fasta").exists():
            return True
        else:
            return False

    def container_call(self, aligner):
        """
        Creates container based query for BLAST align, by adding container run parameters to the base tool
        specific ones.

        Parameters
        ----------
        aligner: str, func
            Tool specific commands.

        Returns
        -------
        str
            Container run command.
        """
        # TODO: Refactor
        if self.db:
            if self._check_db(self.db):
                database = Path(self.db).resolve().parent
            else:
                database = "$HOME/blastdb"
        elif self._check_db(self.subject):
            database = Path(self.subject).resolve().parent
        else:
            raise Exception("No reference file provided, neither a database or subject file")

        # https://github.com/ncbi/blast_plus_docs
        cont_base = f'{self.containerization} run --rm ' \
                    f'-v {database}:/blast/blastdb:ro ' \
                    f'-v {database}:/blast/blastdb_custom:ro ' \
                    f'-v {self.workspace}:/blast/queries:ro ' \
                    f'-v {self.workspace}:/blast/results:rw ' \
                    f'-w /blast {self.image}'
        return " ".join([cont_base, aligner])

    def run_container(self, containerization="docker", image=None, fdb="/blast/blastdb",
                      fquery="/blast/queries", fout="/blast/results"):
        """
        Runs batch align in a local container, this container is only for the align engine.

        Parameters
        ----------
        containerization
            Containerization engine to be used (docker, podman).
        image: str
            Alignment engine container image to be used.
        fdb: str
            Internal container folder for the align databases.
        fquery: str
            Internal container folder for the query file.
        fout: str
            Internal container folder for the align result file.
        """
        if self.db:
            reference = self.db
        else:
            reference = self.subject

        paths = {"db": "/".join([fdb, os.path.basename(reference)]),
                 "query": "/".join([fquery, self.query]),
                 "out": "/".join([fout, self.out])}
        if self.taxidlist:
            paths["taxidlist"] = "/".join([fquery, os.path.basename(self.taxidlist)])

        # Validate type of engine
        if containerization in ["docker", "podman"]:
            self.containerization = containerization
            if image:
                self.image = image
            else:
                self.image = coast_var.default_blast_container

            try:
                base_command = self.generic_aligner(**paths
                                                    )
                subprocess.run(self.container_call(base_command), shell=True, check=True, capture_output=True)
            except subprocess.CalledProcessError as e:
                raise Exception(e.stderr.decode("UTF-8"))
        else:
            raise Exception("Not a compatible containerization engine")

    def merlin_parameters(self):
        """
        Generates a merlin api call parameters.

        Returns
        -------
        str
            Parameters to be written to file.
        """
        query_parameters = []

        for option in ["db", "outfmt", "taxonlist", "taxidlist", "matrix", "evalue", "gapopen", "gapextend"]:
            if getattr(self, option):
                query_parameters.append(
                    f'{self.engine_dict[option].strip("-").strip(" ")} = {os.path.basename(getattr(self, option))}')

        return query_parameters

    def request_merlin(self, files, address, email, raw_output="BlastResults.out", command="BlastP"):
        """
        Call api, fowards request files and confirms returns

        Returns
        -------
        email: str
            Email used for request logging.
        address: str
            Api server base address for request.
        file: list of str
            List of files to be used
        raw_output:
            Merlin given name to the output, to be changed to the coast tool defined one.
        """
        # TODO: check if in the proper network
        multiple_files = [('files', (os.path.basename(file), open(file, 'rb'))) for file in files]
        try:
            r = requests.post(f'{address}/submitRestPlugin/{command}/{email}', files=multiple_files, timeout=2)
            if r.status_code == 201:
                status, timeout, interval = 201, 360000, 10  # in seconds
                attempts = timeout / interval
                attempt = 0
                message = json.loads(r.text)
                submissionID = message["submissionID"]
                while status in [201, 202] and attempt < attempts:
                    time.sleep(10)
                    attempt += 1
                    status_response = requests.get(f'{address}/status/{email}/{submissionID}', timeout=3)
                    status = status_response.status_code
                if status == 200:
                    request.urlretrieve(f'{address}/download/{email}/{submissionID}',
                                        os.path.join(self.workspace, "merlin_run.zip"))
                    unpack_archive(os.path.join(self.workspace, "merlin_run.zip"), extract_dir=self.workspace)
                    result_diamond = hashlib.md5(
                        open(os.path.join(self.workspace, raw_output), 'rb').read()).hexdigest()
                    with open(os.path.join(self.workspace, f'checksum_{raw_output.rsplit(".", 1)[0]}.md5')) as f:
                        first_line = f.readline().strip()
                    if not result_diamond == first_line:
                        raise Exception("Output doesn't match md5 signature provided by the server.")
                    os.rename(os.path.join(self.workspace, raw_output), os.path.join(self.workspace, self.out))
                else:
                    raise Exception(f"{status_response.status_code}: {status_response.reason}")
            else:
                raise Exception(f"{r.status_code}: {r.reason}")
        except requests.exceptions.HTTPError as e:
            return "Error: " + str(e)

    def run_merlin(self, email, address='https://rosalind.di.uminho.pt:6082/'):
        """
        Sends merlin api align request.

        Parameters
        ----------
        email: str
            Email used for request logging.
        address: str
            Api server base address for request.
        """
        protein_path = os.path.join(self.workspace, self.query)
        parameters_file_path = os.path.join(self.workspace, "SubmissionParameters.txt")
        files = [protein_path, parameters_file_path]
        # Generate params
        with open(parameters_file_path, mode='wt', encoding='utf-8') as parameters_file:
            parameters_file.write('\n'.join(self.merlin_parameters()))

        if self.taxidlist:
            files.append(self.taxidlist)
        self.request_merlin(files, address, email, raw_output="BlastResults.out")
